//
//  Implementation of the DSbitFilter class
//
//	AUTHOR:
//	  D. Hinshaw (DH)
//
//	  REFERENCES:
//	    [1] FITS Keyword Conventions in ASC Data Model files, SDS-7.2
// 		J. McDowell
//		http://hea-www.harvard.edu/~jcm/asc/docs/
//
//          [2] ASC Data Model Abstact Design, SDS-3.0
//              J. McDowell
//		http://hea-www.harvard.edu/~jcm/asc/docs/
//	    
//	    [3] Filtering Syntax in the ASC Data Model, SDS-6.0
//	        J. McDowell, A. Fruscione, A. Siemiginowska, B. Joye
//	        http://hea-www.harvard.edu/~jcm/asc/docs/

#include <cstdio>
#include <sstream>
#include <cstring>

#include <Dal.h>
#include <Region.h>
#include <DataSubspace.h>
#include <errstr.h>


DSbitFilter::DSbitFilter(DScomp *parent,
			 const std::string &name, std::string bitString) :
  DSfilter(parent, BitMaskT, name), _onBits(0), _offBits(0) {

  _parseBitString(bitString);
}


DSbitFilter::DSbitFilter(DScomp *parent,
			 const Column *inheritedColumn, 
			 std::string bitString) :
  DSfilter(parent, BitMaskT, inheritedColumn), _onBits(0), _offBits(0) {

    // Check that column is of an integer type
    
    Dal::DataType colType = inheritedColumn->dataType();

    if( colType != Dal::Int8 && 
	colType != Dal::Int16 && 
	colType != Dal::Int32 ){
      errstr << "Column " << inheritedColumn->name() << " must be an integer for bit mask filters"
	     << Sas::error("DSSbitType");
    }

    _parseBitString(bitString);    
}


DSbitFilter::DSbitFilter(DScomp *parent,
			 const std::string &name,
			 vector<unsigned long> onBitMasks,
			 vector<unsigned long> offBitMasks) :
  DSfilter(parent, BitMaskT, name), _onBits(onBitMasks), _offBits(offBitMasks) {}


DSbitFilter::DSbitFilter(DScomp *parent,
			 const Column *inheritedColumn,
			 vector<unsigned long> onBitMasks,
			 vector<unsigned long> offBitMasks) :
  DSfilter(parent, BitMaskT, inheritedColumn), _onBits(onBitMasks), _offBits(offBitMasks) {

    // Check that column is of an integer type
    
    Dal::DataType colType = inheritedColumn->dataType();

    if( colType != Dal::Int8 && 
	colType != Dal::Int16 && 
	colType != Dal::Int32 ){
      errstr << "Column " << inheritedColumn->name() << " must be an integer for bit mask filters"
	     << Sas::error("DSSbitType");
    }
}


void DSbitFilter::write(unsigned compNumber, unsigned filterNumber, 
			Block *toBlock, bool dummy) {
  std::stringstream valueName, value, formName;

  value.setf(ios::showbase);
  if( dummy )
    value << ":";
  else{
    for(unsigned i=0; i<_onBits.size(); i++){
      if(_onBits[i]==0 && _offBits[i]==0) continue;
      if(value.str().size()!=0) value << ',';
      value << 'b';
      
      std::stringstream bits;
      for(unsigned ibit=32; ibit>0; ibit--){
	if(_onBits[i] & 1<<(ibit-1))
	  bits << '1';
	else if(_offBits[i] & 1<<(ibit-1))
	  bits << '0';
	else
	  if(bits.str().size()!=0) bits << 'x';
      }

      value << bits.str();
    }
  }
 
  formName << "DSFORM" << filterNumber;
  if( ! toBlock->hasAttribute(formName.str()) )
    *toBlock->addAttribute(formName.str(), comment+"data type") = "X";
    
  if( compNumber==1 )
    valueName << "DSVAL" << filterNumber;
  else
    valueName << compNumber << "DSVAL" << filterNumber;

  *toBlock->addAttribute(valueName.str(), comment+"value") = value.str();
}


vector<unsigned long> DSbitFilter::onBitMasks() const{
  return _onBits;
}

vector<unsigned long> DSbitFilter::offBitMasks() const{
  return _offBits;
}

std::string DSbitFilter::selectExpression() {
  std::stringstream expression;

  for(unsigned i=0; i<_onBits.size(); i++){
    std::stringstream subExpress;
    if(expression.str().size()>0) expression << " || ";

    if(_offBits[i]!=0)
      subExpress << name() << "&0x" << hex << _offBits[i] <<"==0";

    if(_onBits[i]!=0){
      if(_offBits[i]==0){
	subExpress << name() << "&0x" << hex << _onBits[i] <<"!=0";
      }else{
	for(unsigned ibit=0; ibit<32; ibit++){
	  if(_onBits[i] & 1<<ibit){
	    if(subExpress.str().size()>0) subExpress <<" && ";
	    subExpress << name() << "&0x" << hex << (1<<ibit) <<"!=0";
	  }
	}
      }
    }

    if(subExpress.str().size()>0) expression << subExpress.str();
  }

  return expression.str();
}

 
double DSbitFilter::volume() {
  errstr << "Cannot calculate volume for a bit mask "
	 << "filter.  Zero volumed returned."
         << Sas::warning("BitVol");
  return 0;
}

bool DSbitFilter::operator==(DSfilter &oFilter) const{

  if( name() != oFilter.name()   ||
      type() != oFilter.type()   ||
      units() != oFilter.units() ||
      comps() != oFilter.comps() )
    return false;
      
  if( onBitMasks() != oFilter.onBitMasks()   ||
      offBitMasks() != oFilter.offBitMasks() ) return false;

  return true;
}

void  DSbitFilter::operator!(){
  // For now only support simple filters.
  std::vector<unsigned long> newOn, newOff;

  if(_onBits.size() > 1){
    errstr << "Negation of complex bit strings in the data subspace is not supported."
	   << Sas::error("DssNegBit");
  } 

  if(_onBits[0]==0){
    _onBits[0] = _offBits[0];
    _offBits[0] = 0;
    return;
  }

  

  for(unsigned ibit=0; ibit<32; ibit++){
    if(_onBits[0]&1<<ibit){
      newOn.push_back(0);
      newOff.push_back(1<<ibit);
    }
    if(_offBits[0]&1<<ibit){
      newOn.push_back(1<<ibit);
      newOff.push_back(0);
    }
  }

  _onBits = newOn;
  _offBits = newOff;

  return;
}


void DSbitFilter::_parseBitString(std::string bitString){
  vector<std::string> segments;

  // Only support binary format for now.

  if( bitString[0] != 'b' && ! isdigit(bitString[0]) )
    errstr << "String for bit mask filter (" << bitString 
           << ") is not in a recognizable format." << Sas::error("DssBitString");

  if( isdigit(bitString[0]) ){
    unsigned long temp;
    sscanf(bitString.c_str(), "%lx", &temp);
    _offBits.push_back(temp);
    _onBits.push_back(0);
    return;
  }

  std::string::size_type begin=1;
  std::string::size_type end=bitString.find(",b");
  while(end != std::string::npos){
    segments.push_back(bitString.substr(begin, end-begin+1));
    begin = end+2;
    end=bitString.substr(begin).find(",b");
  }
  segments.push_back(bitString.substr(begin));

  for(unsigned iseg=0; iseg<segments.size(); iseg++){
    unsigned long value=0;
    unsigned nsize = segments[iseg].size();

    for(unsigned i=0; i<nsize; i++){
      if(segments[iseg][i]=='1') value += 1<<(nsize-i-1);
    }
    _onBits.push_back(value);

    value=0;
    for(unsigned i=0; i<nsize; i++){
      if(segments[iseg][i]=='0') value += 1<<(nsize-i-1);
    }
    _offBits.push_back(value);
  }
}
