//
//  Implementation of the DSblockFilter class
//
//	AUTHOR:
//	  D. Hinshaw (DH)
//
//	  REFERENCES:
//	    [1] FITS Keyword Conventions in ASC Data Model files, SDS-7.2
// 		J. McDowell
//		http://hea-www.harvard.edu/~jcm/asc/docs/
//
//          [2] ASC Data Model Abstact Design, SDS-3.0
//              J. McDowell
//		http://hea-www.harvard.edu/~jcm/asc/docs/
//	    
//	    [3] Filtering Syntax in the ASC Data Model, SDS-6.0
//	        J. McDowell, A. Fruscione, A. Siemiginowska, B. Joye
//	        http://hea-www.harvard.edu/~jcm/asc/docs/

#include <sstream>
#include <cstring>
#include <cstdlib>

#include <Dal.h>
#include <errstr.h>
#include <Region.h>
#include <ASCregion.h>
#include <DataSubspace.h>


void DSblockFilter::_openBlock(){
  if(_block!=0) return;
  DataSet * blockset = dataSetServer->open(_parentName, DataSet::Read);
  _block = blockset->block(_blockName);
}


void DSblockFilter::_setBlockSpec(const std::string &blockSpec){
  std::string::size_type icolon = blockSpec.find(":");
  _parentName = blockSpec.substr(0,icolon);
  if(icolon == std::string::npos) _blockName = "0";
  else _blockName = blockSpec.substr(icolon+1, std::string::npos);
}


DSblockFilter::DSblockFilter(DScomp *parent,
			     const FilterType ftype, const std::string &name, 
			     const std::string &blockSpec,
			     const std::string &units ) : 
  DSfilter(parent, ftype, name, units), _block(0) { 

  _setBlockSpec(blockSpec); 
  _openBlock();
  if(_block->hasAttribute("MFORM1")){
    std::string form = _block->attribute("MFORM1")->asString();
    while( form.size() != 0 ){
      std::string::size_type icomma = form.find(",");
      if( icomma == std::string::npos ) icomma = form.size();
      _comps.push_back( form.substr(0,icomma) );
      form.erase(0,icomma+1);
    }
  }
}


DSblockFilter::DSblockFilter(DScomp *parent,
			     const FilterType ftype,
			     const Column *inheritedColumn,
			     const std::string &blockSpec) :
    DSfilter(parent, ftype, inheritedColumn), _block(0) { _setBlockSpec(blockSpec); } 

DSblockFilter::DSblockFilter(DScomp *parent,
			     const FilterType ftype,
			     const std::string &name,
			     const std::string &blockSpec,
			     DSfilterComp &filtercomp) :
  DSfilter(parent, ftype, name, "", filtercomp), _block(0) { _setBlockSpec(blockSpec); }


DSblockFilter::DSblockFilter(DSblockFilter &dsblock) :
  DSfilter(dsblock), _block(0) {
    _parentName = dsblock._parentName;
    _blockName = dsblock._blockName;
    // If the block is stored in a temporary data set, open it here.  Otherwise the
    // temporary set may get deleted when the old DSblockFilter object gets destroyed.
    if( dsblock._block!=0 && dsblock._block->parent()->mode() == DataSet::Temp ) _openBlock();
}
 
 
std::string DSblockFilter::blockSpec() const {
  return _parentName + ":" +  _blockName;
}


void DSblockFilter::write(unsigned compNumber, unsigned filterNumber, 
			  Block *toBlock, bool dummy){
  std::stringstream valueName0;

  std::string value, ref = _blockName;

  valueName0 << "DSVAL" << filterNumber;

  if( type() == MaskT ) 
    value = "IMAGE";
  else
    value = "TABLE";

  if( !toBlock->hasAttribute(valueName0.str()) )
    *toBlock->addAttribute(valueName0.str(), comment+"value") = value;

  if( type() == MaskT || type() == RegionBlockT ){
    DSfilterComp filterComps = comps();
    std::stringstream typeName;
    typeName << "DSTYP" << filterNumber;
    
    *toBlock->attribute(typeName.str()) = toBlock->attribute(typeName.str())->asString() + 
      "(" + filterComps[0] + "," + filterComps[1] + ")";
  }

  std::string refValue;
  if( dummy )
    refValue = ":";
  else{
    _getRefBlockName(compNumber, filterNumber, toBlock, blockSpec(), ref) ;
    if( !toBlock->parent()->hasBlock(ref) ){
      _openBlock();
      toBlock->parent()->add(_block, ref);
    }
    refValue = ":" + ref;
  }

  std::stringstream refName;
  if( compNumber==1 )
    refName << "DSREF" << filterNumber;
  else
    refName << compNumber << "DSREF" << filterNumber;

  *toBlock->addAttribute(refName.str(), comment+"reference") = refValue;

  return;
}


bool DSblockFilter::operator==(DSfilter &oFilter) const{

  if( name() != oFilter.name()   ||
      type() != oFilter.type()   ||
      units() != oFilter.units() ||
      comps() != oFilter.comps() )
    return false;
      
  if( blockSpec() != oFilter.blockSpec() ) return false;

  return true;
}


DSblockFilter::~DSblockFilter() {
  if(_block!=0) dataSetServer->close(_block->parent());
}

