//
//  IMplementation of the DataSubspace class
//
//	AUTHOR:
//	  D. Hinshaw (DH)
//
//	  REFERENCES:
//	    [1] FITS Keyword Conventions in ASC Data Model files, SDS-7.2
// 		J. McDowell
//		http://hea-www.harvard.edu/~jcm/asc/docs/
//
//          [2] ASC Data Model Abstact Design, SDS-3.0
//              J. McDowell
//		http://hea-www.harvard.edu/~jcm/asc/docs/
//
//	    [3] Filtering Syntax in the ASC Data Model, SDS-6.0
//	        J. McDowell, A. Fruscione, A. Siemiginowska, B. Joye
//	        http://hea-www.harvard.edu/~jcm/asc/docs/


#include <cstdio>
#include <sstream>
#include <cstring>
#include <map>


#include <Dal.h>
#include <Region.h>
#include <DataFilter.h>
#include <DataSubspace.h>
#include <errstr.h>


DSS::DSaxes const DSS::noAxes(0);


DataSubspace::DataSubspace() : _compvec(0), _error(false) {}


DataSubspace::DataSubspace(const DataSubspace &dss) {

  for(unsigned icomp=0; icomp<dss.components(); icomp++){
    if( dss.component(icomp)->isNull() ) continue;

    DScomp * comp = new DScomp( *dss.component(icomp) );
    comp->parent(this);
    _compvec.push_back(comp);
  }

  _error = dss.error();

  return;
}


DataSubspace &DataSubspace::operator=(const DataSubspace &dss){
  std::vector<DScomp *>::iterator icomp;
  for(icomp=_compvec.begin(); icomp!=_compvec.end(); icomp++)
    delete *icomp;
  _compvec.clear();

  for(unsigned icomp=0; icomp<dss.components(); icomp++){
    if( dss.component(icomp)->isNull() )continue;

    DScomp * comp = new DScomp( *dss.component(icomp) );
    comp->parent(this);
    _compvec.push_back(comp);
  }

  _error = dss.error();

  return *this;
}


DataSubspace::DataSubspace(Block *inBlock) : _compvec(0){
  inBlock >> *this;
}


DScomp *DataSubspace::addComponent(){
  DScomp *comp = new DScomp(this);
  _compvec.push_back(comp);

  return comp;
}


DScomp *DataSubspace::addComponent(const DScomp &dscomp){
  DScomp *comp = new DScomp(dscomp);
  comp->parent(this);

  _compvec.push_back(comp);
  return comp;
}


void DataSubspace::deleteComponent(unsigned compNumber){
  delete *( _compvec.begin()+compNumber );
  _compvec.erase( _compvec.begin()+compNumber );
}


DScomp *DataSubspace::component(unsigned compNumber) const{
  /*
   * AI: 26/05/2005
   * Check if the _compvec (Vector holding the DSS components) has at least
   * 1 component. If there is no component raises and erro and finish.
   * This fix is needed because emchain's product does not write the STDGTI of
   * a CCD that does not have data. And if you filter this event list
   * writing in the expression the CCD with no data, the _compvec is empty.
   * Wierd, isn't it?
   * Fix SOC-SPR-2583.
   */

  if (_compvec.size() == 0)
    {
      errstr << "DSS vector component is empty. "
	     << Sas::error("DataSubSpace");
    }


  return _compvec[compNumber];
}


unsigned DataSubspace::components() const{
  return _compvec.size();
}


std::string DataSubspace::selectExpression(const DSaxes &axes) const{
  std::vector<DScomp *>::const_iterator icomp;
  std::stringstream expression;

  bool firstcomp = true;
  for(icomp=_compvec.begin(); icomp!=_compvec.end(); icomp++){
    std::string compstr = (*icomp)->selectExpression(axes);
    if( compstr.size() != 0 ){
      if(firstcomp) firstcomp=false;
      else expression << " || ";
      expression << "( " << compstr << " )";
    }
  }

  return expression.str();
}


Range<double> DataSubspace::extent(const std::string &axis) {
  std::vector<DScomp *>::const_iterator icomp;
  std::list<Range<double> > rlist;

  for(icomp=_compvec.begin(); icomp!=_compvec.end(); icomp++)
    rlist.push_back( (*icomp)->extent(axis) );

  return maxRange(rlist);
}

Range<double> DataSubspace::realextent(const std::string &axis) {
  std::vector<DScomp *>::const_iterator icomp;
  std::list<Range<double> > rlist;

  for(icomp=_compvec.begin(); icomp!=_compvec.end(); icomp++)
    rlist.push_back( (*icomp)->realextent(axis) );

  return maxRange(rlist);
}


double DataSubspace::volume(const DSaxes &axes) {
  std::vector<DScomp *>::const_iterator icomp;
  double vol=0;

  for(icomp=_compvec.begin(); icomp!=_compvec.end(); icomp++)
    vol += (*icomp)->volume(axes);

  return vol;
}


std::list<Range<double> > DataSubspace::ranges(const std::string &axis) const{
  std::vector<DScomp *>::const_iterator icomp;
  std::list<Range<double> > r;

  for(icomp=_compvec.begin(); icomp!=_compvec.end(); icomp++){
    std::list<Range<double> > rtemp = (*icomp)->ranges(axis);
    r.merge(rtemp);
  }

  return r;
}


bool DataSubspace::hasFilter(const DSfilter &filter) const{
  std::vector<DScomp *>::const_iterator icomp;

  for(icomp=_compvec.begin(); icomp!=_compvec.end(); icomp++){
    if( (*icomp)->hasFilter(filter) ) return true;
  }

  return false;
}


bool DataSubspace::hasFilter(const std::string &filterName,
		 const FilterType &filterType) const{
  std::vector<DScomp *>::const_iterator icomp;

  for(icomp=_compvec.begin(); icomp!=_compvec.end(); icomp++){
    if( (*icomp)->hasFilter(filterName, filterType) ) return true;
  }

  return false;
}


bool DataSubspace::isGlobal(const DSfilter &filter) const{

  for(unsigned icomp=0; icomp<components(); icomp++){
    DScomp *comp = component(icomp);
    bool global = false;
    for(unsigned ifilter=0; ifilter<comp->filters(); ifilter++)
      if( filter == *comp->filter(ifilter) ){
	global = true;
	break;
      }

    if(!global) return false;
  }

  return true;
}


void DataSubspace::compressRegions() {
  std::vector<DScomp *>::const_iterator icomp;

  for(icomp=_compvec.begin(); icomp!=_compvec.end(); icomp++)
    (*icomp)->compressRegions();


  return;
}

void DataSubspace::operator!(){
  unsigned nComps = _compvec.size();
  if(nComps==0) return;

  std::vector<unsigned> filterCount, compSize;
  DataSubspace oldDss = *this;

  for(unsigned icomp=0; icomp<nComps; icomp++){
    filterCount.push_back(0);
    compSize.push_back(component(icomp)->filters());
  }

  for(int icomp=nComps-1; icomp>=0; icomp--)
    deleteComponent(icomp);
  _compvec.clear();

  DScomp * comp = addComponent();
  while(true){
    for(unsigned icomp=0; icomp<nComps; icomp++)
      comp->addFilter(*oldDss.component(icomp)->filter(filterCount[icomp]));

    filterCount[0]++;
    for(unsigned icomp=0; icomp<nComps-1; icomp++){
      if(filterCount[icomp]==compSize[icomp]){
	filterCount[icomp]=0;
	filterCount[icomp+1]++;
      }else break;
    }
    if(filterCount[nComps-1]==compSize[nComps-1]) break;
    comp = addComponent();
  }

  for(unsigned icomp=0; icomp<components(); icomp++){
    // Need to do this backwards because negating some filters adds more filters to the component
    for(int ifilt=(int)(component(icomp)->filters())-1; ifilt>=0; ifilt--)
      !(*component(icomp)->filter(ifilt));
  }

  return;
}


DataSubspace::~DataSubspace(){
  std::vector<DScomp *>::iterator icomp;

  for(icomp=_compvec.begin(); icomp!=_compvec.end(); icomp++)
    delete *icomp;

  return;
}


Block *operator<<(Block *toBlock, DataSubspace &fromDss){

  // Check to see if the block already contains DSS info,
  // if it does, read it in and merge it with the current DSS,
  // before writing it back out.

  DataSubspace oldDss;
  DataSubspace outDss(fromDss);

  if( fromDss.components() == 0 )  return toBlock;

  // A component with zero filters is considered to be always true.
  // If we have any of these, return

  for(unsigned icomp=0; icomp < outDss.components(); icomp++)
    if( outDss.component(icomp)->filters() == 0 &&
	!outDss.component(icomp)->isNull() ) return toBlock;

  std::string checkType = "DSTYP1";
  if( toBlock->hasAttribute(checkType) ){
    toBlock >> oldDss;
    deleteDSS(toBlock);
    outDss *= oldDss;
  }

  DScomp * comp;
  DSfilter * filter;
  unsigned icomp, ifilt;
  std::string comment = "data subspace descriptor: ";

  std::vector< DSfilter *> attrFilters;
  // There is no reason why there could not be two keywords
  // with the same value. I mean, two filter->name() equals to POS.
  // SOC-SPR-2613
  //std::map<std::string, std::string> compForms;
  std::multimap<std::string, std::string> compForms;


  // Make a first pass thru the DSS, writing type and unit
  // keywords and getting a list of the filters

  for(icomp=0; icomp < outDss.components(); icomp++)
    {
      comp = outDss.component(icomp);
      if(comp->isNull()) continue;
      for(ifilt=0; ifilt < comp->filters(); ifilt++)
	{
	  filter = comp->filter(ifilt);

	  // See if this filter has already been defined,
	  // if not, add it to the list.

	  bool inList = false;
	  for( unsigned i=0; i<attrFilters.size(); i++)
	    if( filter->like(*attrFilters[i]) )
	      {
		inList = true;
		break;
	      }

	  if( !inList )
	    {
	      attrFilters.push_back(filter);

	      // If this is a 2-d filter, store its components
	      if( filter->type() == DSS::RegionT ||
		  filter->type() == DSS::MaskT ||
		  filter->type() == DSS::RegionBlockT )
		{

		  /*
		   * SOC-SPR-2613
		   * It is possible than have an expression with more than one identical
		   * filter->name(). If this is the case, more than one data subspace will
		   * be written to the header. Due to this fact, I have decrease this error
		   * message to a warning message.
		   */

		  if( compForms.find(filter->name()) != compForms.end() )
		    errstr << "Two 2-d filters with the name " << filter->name()
			   << " have been found to different components. "
			   << "Cannot write data subspace."
			   << Sas::warning("DssCompForm");


		  DSS::DSfilterComp comps = filter->comps();
		  /*
		   * SOC-SPR-2613
		   * compForms is defined as a multimap, in stead of, a map.
		   * compForms[filter->name()] = comps[0] + "," + comps[1];
		   */
		  compForms.insert(make_pair(filter->name(), comps[0] + "," + comps[1]));
		}



	      // Check for a name clash between this filter and a column

	      if( toBlock->type() == Block::TableT ){
		Table * toTable = dynamic_cast<Table *>(toBlock);
		if( toTable->hasColumn(filter->name()) &&
		    ( filter->type() == DSS::RegionT ||
		      filter->type() == DSS::MaskT ||
		      filter->type() == DSS::RegionBlockT ||
		      filter->units() != toTable->column(filter->name())->units() )
		    )
		  {
		    cout << "units, block, col: " << filter->units() <<" "
			 << toTable->column(filter->name())->units() << endl;

		    errstr << "Data subspace filter name and column name "
			   << filter->name()
			   << " are not consistent.  Cannot write data subspace "
			   << "to table " << toTable->name() << Sas::error("ColumnClash");
		  }
	      }

	      std::stringstream name0;
	      name0 << "DSTYP" << attrFilters.size();
	      *toBlock->addAttribute(name0.str(), comment+"name") = filter->name();

	      if( filter->units() != "" )
		{
		  std::stringstream unit0;
		  unit0 << "DSUNI" << attrFilters.size();
		  *toBlock->addAttribute(unit0.str(), comment+"units") =
		    filter->units();
		}
	    }
	}
    }

  // Now make a second pass thru the DSS, writing value and
  // reference keywords, as appopriate for the type of filter.

  DScomp * comp0 = outDss.component(0);
  icomp=0;
  for(unsigned i=0; i < outDss.components(); i++)
    {
      comp = outDss.component(icomp);
      if(comp->isNull()) continue;
      for( unsigned ifilt=0; ifilt<attrFilters.size(); ifilt++)
	{
	  if((icomp >= 9 && ifilt >= 9)  || icomp >= 99 || ifilt >= 99 )
	    {
	      errstr << "The data subspace specification has become too complex"
		     << "(component " << icomp+1 << ", filter " << ifilt+1 << ") "
		     << "to be able to write using the FITS keyword conventions. "
		     << "The data subspace information stored in the FITS files will "
		     << "be incomplete." << Sas::warning("DssFitsWrite");
	      fromDss.setError();
	      continue;
	    }

      if( comp->hasFilter(*attrFilters[ifilt]) )
	{
	  if( icomp==0 ||
	      !comp0->hasFilter(*attrFilters[ifilt]) ||
	      *comp0->filter(*attrFilters[ifilt]) !=
	      *comp->filter(*attrFilters[ifilt]) )

	    comp->filter(*attrFilters[ifilt])
	      ->write(icomp+1, ifilt+1, toBlock);

	}
      else if( icomp==0 || comp0->hasFilter(*attrFilters[ifilt]) )
	{

	  // No filter for this component, so write a dummy
	  attrFilters[ifilt]
	    ->write(icomp+1, ifilt+1, toBlock, true);
	}
	}
      icomp++;
    }

  // Write out the MTYPE keywords
  /*
   * SOC-SPR-2613
   * We define the variable as a multimap in order to have the possibility
   * of write several kwywords with the same value (mainly POS)
   *  std::map<std::string, std::string>::iterator form;
   */
  std::multimap<std::string, std::string>::iterator form;
  int iform=0;

  for(form=compForms.begin(); form!=compForms.end(); form++)
    {
      iform++;
      stringstream typeName, formName;
      typeName << "MTYPE" << iform;
      formName << "MFORM" << iform;
      *toBlock->addAttribute(typeName.str(), "DM Keyword: Descriptor name")
	= form->first;
      *toBlock->addAttribute(formName.str())
	= form->second;
    }

  return toBlock;
}


Block *operator>>(Block *fromBlock, DataSubspace &toDss){

  DataSubspace inDss;

  // Count the number of DSS filters and components in the block

  unsigned nFilters;

  for(nFilters=0; ; nFilters++){
    std::stringstream typeName;
    typeName << "DSTYP" << nFilters+1;
    if( ! fromBlock->hasAttribute(typeName.str()) ) break;
  }

  if( nFilters==0 ) return fromBlock;

  unsigned nComps;

  for(nComps=1; ; nComps++){
    inDss.addComponent();

    bool values = false;
    for(unsigned ifilt=0; ifilt<nFilters; ifilt++){
      std::stringstream valName, refName;
      valName << nComps+1 << "DSVAL" << ifilt+1;
      refName << nComps+1 << "DSREF" << ifilt+1;
      if( fromBlock->hasAttribute(valName.str()) ||
 	  fromBlock->hasAttribute(refName.str()) )
 	values = true;
    }

    if( !values )break;
  }

  // Now loop thru again, adding filters as needed

  std::map<std::string, std::string> filterBlocks;

  for(unsigned ifilt=1; ifilt<=nFilters; ifilt++){
    std::stringstream typeName, unitName;
    typeName << "DSTYP" << ifilt;

    bool hasFiltComp = false;
    std::string compVal;
    std::string typeVal = fromBlock->attribute(typeName.str())->asString();
    std::string::size_type iparse = typeVal.find("(");
    if( iparse != std::string::npos ){
      hasFiltComp = true;
      compVal = typeVal.substr(iparse);
      typeVal = typeVal.substr(0,iparse);
    }

    unitName << "DSUNI" << ifilt;
    std::string unitVal;
    if( fromBlock->hasAttribute(unitName.str()) )
      unitVal = fromBlock->attribute(unitName.str())->asString();
    else
      unitVal = "";

    // Is there a column associated with this filter?

    Column * inheritCol = 0;
    bool inherit = false;
    if( fromBlock->type() == Block::TableT ){
      Table * fromTable = dynamic_cast<Table *>(fromBlock);
      if( fromTable->hasColumn(typeVal) ){
	inherit = true;
	inheritCol = fromTable->column(typeVal);
      }
    }

    for(unsigned icomp=0; icomp<nComps; icomp++){
      std::stringstream valName0;
      valName0 << "DSVAL" << ifilt;
      std::string value;

      // Get the filter value

      if( icomp==0 ){
	value = fromBlock->attribute(valName0.str())->asString();
      }else{
	std::stringstream valName;
	valName << icomp+1 << "DSVAL" << ifilt;
	if( fromBlock->hasAttribute(valName.str()) )
	  value = fromBlock->attribute(valName.str())->asString();
	else
	  value = fromBlock->attribute(valName0.str())->asString();
      }

      // Determine filter type.  Do nothing if empty filter

      if( value==":" ) continue;

      // Block filter

      if( value=="IMAGE" || value=="TABLE" ){
	std::stringstream refName0;
	refName0 << "DSREF" << ifilt;
	std::string ref;

	if( icomp==0 ){
	  ref = fromBlock->attribute(refName0.str())->asString();
	}else{
	  std::stringstream refName;
	  refName << icomp+1 << "DSREF" << ifilt;
	  if( fromBlock->hasAttribute(refName.str()) )
	    ref = fromBlock->attribute(refName.str())->asString();
	  else
	    ref = fromBlock->attribute(refName0.str())->asString();
	}

	if( ref==":" ) continue;
	ref.erase(0,1);

	if( filterBlocks[ref]=="" )
	  filterBlocks[ref] = fromBlock->parent()->name() + ":" + ref;

	if(inherit){
	  inDss.component(icomp)->addBlockFilter(inheritCol, filterBlocks[ref]);
	  continue;
	}

	if( hasFiltComp ){
	  DSS::DSfilterComp dscomps;
	  std::string::size_type icomma = compVal.find(",");
	  std::string::size_type oparen = compVal.find("(");
	  std::string::size_type cparen = compVal.find(")");

	  dscomps.push_back(compVal.substr(oparen+1,icomma-oparen-1));
	  dscomps.push_back(compVal.substr(icomma+1,cparen-icomma-1));

	  inDss.component(icomp)->addBlockFilter(typeVal, filterBlocks[ref], dscomps);
	  continue;
	}

	inDss.component(icomp)->addBlockFilter(typeVal, filterBlocks[ref], unitVal);
	continue;
      }

      // Bit filter

      std::stringstream formName;
      formName << "DSFORM" << ifilt;

      if( fromBlock->hasAttribute(formName.str()) &&
	  fromBlock->attribute(formName.str())->asString()=="X" ){

	if(inherit)
	  inDss.component(icomp)->addBitFilter(inheritCol, value);
	else
	  inDss.component(icomp)->addBitFilter(typeVal, value);

	continue;
      }

      // Range filter

      char firstVal = *value.substr(0,1).data();
      if( isdigit(firstVal) || firstVal=='[' ||
	  firstVal=='(' || firstVal=='.' || firstVal==':' || firstVal=='-'){

	bool isInt = false;
	if( inherit )
	  isInt = inheritCol->dataType() == Dal::Int8 ||
	    inheritCol->dataType() == Dal::Int16 ||
	    inheritCol->dataType() == Dal::Int32;
	else if( fromBlock->hasAttribute(formName.str()) )
	  isInt = fromBlock->attribute(formName.str())->asString()=="B" ||
	    fromBlock->attribute(formName.str())->asString()=="I" ||
	    fromBlock->attribute(formName.str())->asString()=="J";
	else
	  isInt = value.find(".") == std::string::npos;

	if( isInt ){
	  std::list<Range<long> > inRanges = parseRanges<long>(value);
	  if(inherit) inDss.component(icomp)->addRangeFilter(inheritCol, inRanges);
	  else inDss.component(icomp)->addRangeFilter(typeVal, inRanges, unitVal);
	}else{
	  std::list<Range<double> > inRanges = parseRanges<double>(value);
	  if(inherit) inDss.component(icomp)->addRangeFilter(inheritCol, inRanges);
	  else inDss.component(icomp)->addRangeFilter(typeVal, inRanges, unitVal);
	}

	continue;
      }

      // Region filter - not supported

      errstr << "A data subspace filter with a region specification "
	     << "has been found.  This type of filter is not supported."
	     << Sas::error("BadRegionFilter");
    }

  }

  // If the DSS object is not empty, need to merge new DSS info with old.

  if( toDss.components() != 0 )
    toDss *= inDss;
  else
    toDss = inDss;

  return fromBlock;
}

// Perform union of two DSS's.  This simply involves ANDing together each of
// the components of one subspace with each of the components of the other.

DataSubspace &operator*=(DataSubspace &dss1, const DataSubspace &dss2){
  dss1 = dss1 * dss2;
  return dss1;
}


DataSubspace operator*(const DataSubspace &dss1, const DataSubspace &dss2){
  DataSubspace dssResult;

  for(unsigned icomp1=0; icomp1<dss1.components(); icomp1++)
    for(unsigned icomp2=0; icomp2<dss2.components(); icomp2++){
      DScomp comp = *dss1.component(icomp1) * *dss2.component(icomp2);
      if( !comp.isNull() )
	dssResult.addComponent(comp);
    }

  return dssResult;
}


// Delete the DSS specification from a block, not including referenced extensions.
void deleteDSS(Block *block){

  // Count the number of DSS filters and components in the block

  unsigned nFilters;

  for(nFilters=0; ; nFilters++){
    std::stringstream typeName;
    typeName << "DSTYP" << nFilters+1;
    if( ! block->hasAttribute(typeName.str()) ) break;

    block->deleteAttribute(typeName.str());

    std::stringstream unitName;
    unitName << "DSUNI" << nFilters+1;
    if( block->hasAttribute(unitName.str()) ) block->deleteAttribute(unitName.str());

    std::stringstream cmpName;
    cmpName << "DSCPT" << nFilters+1;
    if( block->hasAttribute(cmpName.str()) ) block->deleteAttribute(cmpName.str());

    std::stringstream formName;
    formName << "DSFORM" << nFilters+1;
    if( block->hasAttribute(formName.str()) ) block->deleteAttribute(formName.str());
  }

  if( nFilters==0 ) return;

  for(unsigned nComps=1; ; nComps++){
    bool values = false;
    for(unsigned ifilt=0; ifilt<nFilters; ifilt++){
      std::stringstream valName, refName;
      if( nComps==1 ){
	valName << "DSVAL" << ifilt+1;
	refName << "DSREF" << ifilt+1;
      }else{
	valName << nComps << "DSVAL" << ifilt+1;
	refName << nComps << "DSREF" << ifilt+1;
      }
      if( block->hasAttribute(valName.str()) ){
	block->deleteAttribute(valName.str());
	values = true;
      }
      if( block->hasAttribute(refName.str()) ){
	block->deleteAttribute(refName.str());
	values = true;
      }
    }
    if(!values) break;
  }

}


void cleanDSS(const Table *tab, DataSubspace &dss){
  DSS::DSaxes cols;

  for(unsigned icol=0; icol<tab->columns(); icol++)
    cols.push_back(tab->column(icol)->name());

  for(unsigned icomp=dss.components(); icomp!=0; icomp--){
    DScomp * comp = dss.component(icomp-1);
    DSS::DSaxes axes = cols;

    bool infiniteRange = false;
    for(unsigned ifilter=0; ifilter<comp->filters(); ifilter++){
      DSfilter * filter = comp->filter(ifilter);
      if( filter->type() == DSS::RangeT ){
	std::list<Range<double> > ranges = filter->ranges();
	if( ranges.size() == 1 ){
	  double lower, upper;
	  RangeBound lowerB, upperB;
	  ranges.begin()->values(lower, upper, lowerB, upperB);
	  if( lowerB==Closed && upperB==Closed &&
	      lower==num_limits<double>::max && upper==num_limits<double>::max ){
	    infiniteRange = true;
	    break;
	  }
	}
      }

      if( filter->type() != DSS::RegionT      &&
	  filter->type() != DSS::MaskT        &&
	  filter->type() != DSS::RegionBlockT ) continue;
      DSS::DSfilterComp filtComp = filter->comps();
      if( tab->hasColumn(filtComp[0]) && tab->hasColumn(filtComp[1]) )
	axes.push_back(filter->name());
    }

    if(infiniteRange) dss.deleteComponent(icomp-1);
    else{
      DataFilter filter(comp->selectExpression(axes));
      std::vector<DataFilter::RowRange> rowranges;
      if( filter.doSelect(*tab, rowranges)==0 )	dss.deleteComponent(icomp-1);
    }
  }
}
